
//
//  Instrument.swift
//
//  Created by David Skulina on 27/09/2016.
//  Copyright © 2016 Skoogmusic Ltd. All rights reserved.
//
import Foundation
import AudioToolbox
import CoreAudio

public enum SynthType {
    case fmModulator
    case sineWave
    case solarWind
    case candyBee
    case superSaw
    case acid
    var string: String {
        switch self {
        case .fmModulator:
            return "Soundfonts/FM Modulator"
        case .sineWave:
            return "Soundfonts/Sine Wave"
        case .solarWind:
            return "Soundfonts/Solar Wind"
        case .candyBee:
            return "Soundfonts/Candy Bee"
        case .superSaw:
            return "Soundfonts/Super Saw 1"
        case .acid:
            return "Soundfonts/Acid SQ Neutral"
        }
    }
}

/// # A Core Audio MIDISynth `AudioUnit` example.
/// This will add a polyphonic `kAudioUnitSubType_MIDISynth` audio unit to the `AUGraph`.
///
/// - author: Gene De Lisa
/// - copyright: 2016 Gene De Lisa
/// - date: February 2016
public class Instrument : NSObject {
    
    var processingGraph: AUGraph?
    var midisynthNode   = AUNode()
    var ioNode          = AUNode()
    var midisynthUnit: AudioUnit?
    
    var ioUnit: AudioUnit?
    var isPlaying:Bool
    
    
    /// Harp
    let patch1          = UInt32(12)
    

    public var type: SynthType = .sineWave
    
    
    /// Initialize.
    /// set up the graph, load a sound font into the synth, create a sequence, create a player for
    /// that sequence, and start the graph.
    public override init() {
        self.processingGraph	= nil
        self.midisynthUnit		= nil
        self.ioUnit				= nil
        self.isPlaying			= false
        super.init()
        
        
        augraphSetup()
//        loadMIDISynthSoundFont(type: .solarWind)
        loadPatches()
        startGraph()
    }
    
    /// Create the `AUGraph`, the nodes and units, then wire them together.
    func augraphSetup() {
        
        NewAUGraph(&processingGraph)
        
        createIONode()
        
        createSynthNode()
        
        // now do the wiring. The graph needs to be open before you call AUGraphNodeInfo
        AUGraphOpen(self.processingGraph!)
        AUGraphNodeInfo(self.processingGraph!, self.midisynthNode, nil, &midisynthUnit)
        AUGraphNodeInfo(self.processingGraph!, self.ioNode, nil, &ioUnit)
        
        let synthOutputElement:AudioUnitElement = 0
        let ioUnitInputElement:AudioUnitElement = 0
        
        AUGraphConnectNodeInput(self.processingGraph!,
                                self.midisynthNode, synthOutputElement, // srcnode, SourceOutputNumber
            self.ioNode, ioUnitInputElement) // destnode, DestInputNumber
    }
    
    /// Create the Output Node and add it to the `AUGraph`.
    func createIONode() {
        var cd = AudioComponentDescription(
            componentType: OSType(kAudioUnitType_Output),
            componentSubType: OSType(kAudioUnitSubType_RemoteIO),
            componentManufacturer: OSType(kAudioUnitManufacturer_Apple),
            componentFlags: 0,componentFlagsMask: 0)
        AUGraphAddNode(self.processingGraph!, &cd, &ioNode)
    }
    
    /// Create the Synth Node and add it to the `AUGraph`.
    func createSynthNode() {
        var cd = AudioComponentDescription(
            componentType: OSType(kAudioUnitType_MusicDevice),
            componentSubType: OSType(kAudioUnitSubType_MIDISynth),
            componentManufacturer: OSType(kAudioUnitManufacturer_Apple),
            componentFlags: 0,componentFlagsMask: 0)
        AUGraphAddNode(self.processingGraph!, &cd, &midisynthNode)
    }
    
    
    /// This will load the default sound font and set the synth unit's property.
    /// - postcondition: `self.midisynthUnit` will have it's sound font url set.
    public func loadMIDISynthSoundFont(type: SynthType)  {
        
        if var bankURL = Bundle.main.url(forResource: type.string, withExtension: "sf2")  {
            
            AudioUnitSetProperty(
                self.midisynthUnit!,
                AudioUnitPropertyID(kMusicDeviceProperty_SoundBankURL),
                AudioUnitScope(kAudioUnitScope_Global),
                0,
                &bankURL,
                UInt32(MemoryLayout<URL>.size))
            
        } else {
            print("Could not load sound font")
        }
        print("loaded sound font")
    }
    
    
    /// Pre-load the patches you will use.
    ///
    /// Turn on `kAUMIDISynthProperty_EnablePreload` so the midisynth will load the patch data from the file into memory.
    /// You load the patches first before playing a sequence or sending messages.
    /// Then you turn `kAUMIDISynthProperty_EnablePreload` off. It is now in a state where it will respond to MIDI program
    /// change messages and switch to the already cached instrument data.
    ///
    /// - precondition: the graph must be initialized
    ///
    /// [Doug's post](http://prod.lists.apple.com/archives/coreaudio-api/2016/Jan/msg00018.html)
    func loadPatches() {
        
        var enabled = UInt32(1)
        
        AudioUnitSetProperty(
            self.midisynthUnit!,
            AudioUnitPropertyID(kAUMIDISynthProperty_EnablePreload),
            AudioUnitScope(kAudioUnitScope_Global),
            0,
            &enabled,
            UInt32(MemoryLayout<UInt32>.size))
        
        //        let bankSelectCommand = UInt32(0xB0 | 0)
        //        status = MusicDeviceMIDIEvent(self.midisynthUnit, bankSelectCommand, 0, 0, 0)
        
        for n : UInt32 in 0...4 {
            let pcCommand = UInt32(0xC0 | n)
            MusicDeviceMIDIEvent(self.midisynthUnit!, pcCommand, patch1, 0, n)
        }
        
        enabled = UInt32(0)
        AudioUnitSetProperty(
            self.midisynthUnit!,
            AudioUnitPropertyID(kAUMIDISynthProperty_EnablePreload),
            AudioUnitScope(kAudioUnitScope_Global),
            0,
            &enabled,
            UInt32(MemoryLayout<UInt32>.size))
        
        // at this point the patches are loaded. You still have to send a program change at "play time" for the synth
        // to switch to that patch
    }
    
    
    /// Starts the `AUGraph`
    func startGraph() {
        var outIsInitialized = DarwinBoolean(false)
        AUGraphIsInitialized(self.processingGraph!, &outIsInitialized)
        AUGraphInitialize(self.processingGraph!)
        
        var isRunning = DarwinBoolean(false)
        AUGraphIsRunning(self.processingGraph!, &isRunning)
        
        AUGraphStart(self.processingGraph!)
        
        self.isPlaying = true
    }
    
    public func noteOn(noteNum:UInt, velocity:UInt32, channel: UInt32)    {
        // note on command
        let noteCommand = UInt32(0x90 | channel)
        MusicDeviceMIDIEvent(self.midisynthUnit!, noteCommand, UInt32(noteNum), velocity, channel)
        
    }
    
    public func noteOff(noteNum:UInt, channel: UInt32)    {
        // note off command
        let noteCommand = UInt32(0x80 | channel)
        MusicDeviceMIDIEvent(self.midisynthUnit!, noteCommand, UInt32(noteNum), 0, 0)
    }
    
    public func channelPressure(pressure:UInt32, channel: UInt32)    {
        // channel pressure command
        let noteCommand = UInt32(0xA0 | channel)
        MusicDeviceMIDIEvent(self.midisynthUnit!, noteCommand, pressure, 0, channel)
    }
	
	public func afterTouch(note: UInt, touch:UInt32, channel: UInt32)    {
		// aftertouch command
		let noteCommand = UInt32(0xA0 | channel)
		MusicDeviceMIDIEvent(self.midisynthUnit!, noteCommand, UInt32(note), touch, channel)
	}
	
    public func pitchBend(bend:UInt32, channel: UInt32)    {
        // pitchbend command
        let lsb = UInt8(bend & 0xFF)
        let msb = UInt8((bend >> 7) & 0xFF)
        let noteCommand = UInt32(0xE0 | channel)
        MusicDeviceMIDIEvent(self.midisynthUnit!, noteCommand, UInt32(lsb), UInt32(msb), channel)
    }
    
    public func pitchRange(range:UInt32, channel: UInt32){
        let noteCommand = UInt32(0xB0 | channel)
        MusicDeviceMIDIEvent(self.midisynthUnit!, noteCommand, 101, 0, channel)
        MusicDeviceMIDIEvent(self.midisynthUnit!, noteCommand, 100, 0, channel)
        MusicDeviceMIDIEvent(self.midisynthUnit!, noteCommand, 6, range, channel)
        MusicDeviceMIDIEvent(self.midisynthUnit!, noteCommand, 38, 0, channel)
    }
    
    public func volume(value:UInt32, channel: UInt32)    {
        // volume change command
        let noteCommand = UInt32(0xB0 | channel)
        MusicDeviceMIDIEvent(self.midisynthUnit!, noteCommand, 7, value, channel)
    }
    
    public func modulate(value:UInt32, channel: UInt32)    {
        // volume change command
        let noteCommand = UInt32(0xB0 | channel)
        MusicDeviceMIDIEvent(self.midisynthUnit!, noteCommand, 1, value, channel)
    }
    
    public func expression(value:UInt32, channel: UInt32)    {
        // volume change command
        let noteCommand = UInt32(0xB0 | channel)
        MusicDeviceMIDIEvent(self.midisynthUnit!, noteCommand, 11, value, channel)
    }
    
    public func pan(value:UInt32, channel: UInt32)    {
        // volume change command
        let noteCommand = UInt32(0xB0 | channel)
        MusicDeviceMIDIEvent(self.midisynthUnit!, noteCommand, 10, value, channel)
    }
    
}
